import requests
import os
import hashlib
import time
import pprint


"""Test Pandarus remote.

Need to have:

* Flask app running. Do something like (from the pandarus_remote main directory):
    export FLASK_APP=pandarus_remote/__init__.py
    export FLASK_DEBUG=1
    flask run

* Rq worker (anywhere, but in correct virtualenv):
    rq worker

* Redis
    redis-server

"""

def sha256(filepath, blocksize=65536):
    """Generate SHA 256 hash for file at `filepath`"""
    hasher = hashlib.sha256()
    fo = open(filepath, 'rb')
    buf = fo.read(blocksize)
    while len(buf) > 0:
        hasher.update(buf)
        buf = fo.read(blocksize)
    return hasher.hexdigest()

url = 'http://127.0.0.1:5000'
# url = 'https://pandarus.brightwaylca.org'

print("Ping!")
print(requests.get(url).text)

print("Checking basic data upload")

os.chdir("data")

fp = 'grid.geojson'
print(fp)
data = {
    'field': 'name',
    'name': os.path.basename(fp),
    'sha256': sha256(fp),
    'layer': '',
    'band': ''
}
resp = requests.post(url + '/upload', data=data, files={'file': open(fp, 'rb')})
assert resp.status_code == 200

fp = 'square.geojson'
print(fp)
data = {
    'field': 'name',
    'name': os.path.basename(fp),
    'sha256': sha256(fp),
    'layer': '',
    'band': ''
}
resp = requests.post(url + '/upload', data=data, files={'file': open(fp, 'rb')})
assert resp.status_code == 200

fp = 'range.tif'
print(fp)

data = {
    'field': '',
    'name': os.path.basename(fp),
    'sha256': sha256(fp),
    'layer': '',
    'band': '1'
}

resp = requests.post(url + '/upload', data=data, files={'file': open(fp, 'rb')})
assert resp.status_code == 200

print('Catalog:')
pprint.pprint(requests.get(url + '/catalog').json())

print("Check intersections")
data = {
    'first': sha256('square.geojson'),
    'second': sha256('grid.geojson')
}
resp = requests.post(url + "/calculate-intersection", data=data)
assert resp.status_code == 200

time.sleep(3)

print("Check remaining")
resp = requests.post(url + "/calculate-remaining", data=data)
assert resp.status_code == 200

print("Check raster stats")
data = {
    'raster': sha256('range.tif'),
    'vector': sha256('grid.geojson')
}

resp = requests.post(url + "/calculate-rasterstats", data=data)
assert resp.status_code == 200

time.sleep(3)

print('Catalog:')
pprint.pprint(requests.get(url + '/catalog').json())

print("Retrieving data")

print("Raster stats")
data = {
    'raster': sha256('range.tif'),
    'vector': sha256('grid.geojson')
}
resp = requests.post(url + "/rasterstats", data=data)
assert resp.status_code == 200

print("Intersection")
data = {
    'first': sha256('square.geojson'),
    'second': sha256('grid.geojson')
}
resp = requests.post(url + "/intersection", data=data)
assert resp.status_code == 200

resp = requests.post(url + "/intersection-file", data=data)
assert resp.status_code == 200

print("Remaining")
resp = requests.post(url + "/remaining", data=data)
assert resp.status_code == 200



from pandarus.rasters import *
import os

dirpath = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))
grid = os.path.join(dirpath, "grid.geojson")
lines = os.path.join(dirpath, "lines.geojson")
points = os.path.join(dirpath, "points.geojson")
range_raster = os.path.join(dirpath, "range.tif")

def test_rasterstats_vector():
    result = list(gen_zonal_stats(grid, range_raster))
    assert result == [
        {'count': 10.200000762939453, 'min': 30.0, 'mean': 38.32352828979492, 'max': 47.0},
        {'count': 10.199999809265137, 'min': 0.0, 'mean': 8.323529243469238, 'max': 17.0},
        {'count': 10.0, 'min': 32.0, 'mean': 40.70000076293945, 'max': 49.0},
        {'count': 10.0, 'min': 2.0, 'mean': 10.699999809265137, 'max': 19.0},
    ]

def test_rasterstats_lines():
    result = list(gen_zonal_stats(lines, range_raster))
    assert result == [
        {'max': 36.0, 'min': 11.0, 'mean': 18.450448989868164, 'count': 0.2775000035762787},
        {'max': 38.0, 'min': 33.0, 'mean': 34.666664123535156, 'count': 0.07500000298023224}
    ]

def test_rasterstats_points():
    result = list(gen_zonal_stats(points, range_raster))
    assert result == [
        {'min': 30.0, 'max': 42.0, 'count': 9, 'mean': 36.0},
        {'min': 7.0, 'max': 19.0, 'count': 9, 'mean': 13.0}
    ]



from pandarus import Map
from pandarus.projection import project, WGS84, MOLLWEIDE
from pandarus.geometry import (
    clean,
    get_intersection as _get_intersection,
    get_measure,
    get_remaining,
    IncompatibleTypes,
    recursive_geom_finder,
)
from shapely.geometry import (
    GeometryCollection,
    LinearRing,
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
    mapping,
)
import numpy as np
import os
import pytest

dirpath = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))
grid = os.path.join(dirpath, "grid.geojson")

to_mollweide = lambda x: project(x, WGS84, MOLLWEIDE)


def get_intersection(*args, **kwargs):
    """Unwrap ``get_intersection`` result dictionaries to give geometries in GeoJSON"""
    dct = _get_intersection(*args, **kwargs)
    for v in dct.values():
        if 'geom' in v:
            v['geom'] = mapping(v['geom'])
    return dct


# get_intersection

def test_no_return_geoms():
    mp = Point((0.5, 1))
    expected = {
        0: {'measure': 1},
        1: {'measure': 1}
    }
    result = get_intersection(mp, 'point', Map(grid, 'name'), (0, 1, 2, 3), to_meters=False, return_geoms=False)
    print(result)
    assert result == expected

def test_recurse_geometry_collection():
    mp = MultiPoint([(0.5, 0.5), (0.5, 1), (1, 1), (1.5, 1.5)])
    gc = GeometryCollection([mp])
    assert recursive_geom_finder(gc, 'point').wkt == mp.wkt

    mp = MultiPoint([(0.5, 0.5), (0.5, 1), (1, 1), (1.5, 1.5)])
    gc = GeometryCollection([mp])
    assert recursive_geom_finder(gc, 'line') is None


# Points

def test_single_point():
    mp = Point((0.5, 1))
    expected = {
        0: {
            'geom': {'type': 'MultiPoint', 'coordinates': ((0.5, 1.0),)},
            'measure': 1},
        1: {
            'geom': {'type': 'MultiPoint', 'coordinates': ((0.5, 1.0),)},
            'measure': 1}
        }
    result = get_intersection(mp, 'point', Map(grid, 'name'), (0, 1, 2, 3))
    assert result == expected

    expected = {0: {'geom': {'type': 'MultiPoint', 'coordinates': ((0.5, 1.0),)}, 'measure': 1}}
    assert get_intersection(mp, 'point', Map(grid, 'name'), (0, 2)) == expected

def test_multi_point():
    mp = MultiPoint([(0.5, 0.5), (0.5, 1), (1, 1), (1.5, 1.5)])
    expected = {
        0: {
            'geom': {'coordinates': ((0.5, 0.5), (0.5, 1.0), (1.0, 1.0)), 'type': 'MultiPoint'},
            'measure': 3},
        1: {
            'geom': {'coordinates': ((0.5, 1.0), (1.0, 1.0)), 'type': 'MultiPoint'},
            'measure': 2},
        2: {
            'geom': {'coordinates': ((1.0, 1.0),), 'type': 'MultiPoint'},
            'measure': 1},
        3: {
            'geom': {'coordinates': ((1.0, 1.0), (1.5, 1.5)), 'type': 'MultiPoint'},
            'measure': 2}
    }
    assert get_intersection(mp, 'point', Map(grid, 'name'), (0, 1, 2, 3)) == expected

def test_point_geometry_collection():
    mp = GeometryCollection([MultiPoint([(0.5, 0.5), (0.5, 1), (1, 1), (1.5, 1.5)])])
    expected = {
        0: {
            'geom': {'coordinates': ((0.5, 0.5), (0.5, 1.0), (1.0, 1.0)), 'type': 'MultiPoint'},
            'measure': 3},
        1: {
            'geom': {'coordinates': ((0.5, 1.0), (1.0, 1.0)), 'type': 'MultiPoint'},
            'measure': 2},
        2: {
            'geom': {'coordinates': ((1.0, 1.0),), 'type': 'MultiPoint'},
            'measure': 1},
        3: {
            'geom': {'coordinates': ((1.0, 1.0), (1.5, 1.5)), 'type': 'MultiPoint'},
            'measure': 2}
    }
    assert get_intersection(mp, 'point', Map(grid, 'name'), (0, 1, 2, 3)) == expected

def test_point_wrong_geometry():
    ls = LineString([(0.5, 0.5), (1.5, 0.5)])
    assert get_intersection(ls, 'point', Map(grid, 'name'), (0, 1, 2, 3)) == {}

# Lines

def test_line_string():
    ls = LineString([(0.5, 0.5), (1.5, 0.5)])
    expected = {
        0: {
            'measure': 0.5,
            'geom': {'coordinates': (((0.5, 0.5), (1.0, 0.5)),), 'type': 'MultiLineString'}},
        2: {
            'measure': 0.5,
            'geom': {'coordinates': (((1.0, 0.5), (1.5, 0.5)),), 'type': 'MultiLineString'}}
    }
    result = get_intersection(ls, 'line', Map(grid, 'name'), (0, 1, 2, 3), to_meters=False)
    assert result == expected

    expected = {0: {
        'geom': {'type': 'MultiLineString', 'coordinates': (((0.5, 0.5), (1.0, 0.5)),)},
        'measure': 0.5
    }}
    result = get_intersection(ls, 'line', Map(grid, 'name'), (0, 1), to_meters=False)
    print(result)
    assert result == expected

def test_multi_line_string():
    ls = MultiLineString([[(0.5, 0.5), (1.5, 0.5)]])
    expected = {
        0: {
            'measure': 0.5,
            'geom': {'type': 'MultiLineString', 'coordinates': (((0.5, 0.5), (1.0, 0.5)),)}},
        2: {
            'measure': 0.5,
            'geom': {'type': 'MultiLineString', 'coordinates': (((1.0, 0.5), (1.5, 0.5)),)}}
    }
    assert get_intersection(ls, 'line', Map(grid, 'name'), (0, 1, 2, 3), to_meters=False) == expected

    expected = {0: {
        'measure': 0.5,
        'geom': {'coordinates': (((0.5, 0.5), (1.0, 0.5)),), 'type': 'MultiLineString'}
    }}
    assert get_intersection(ls, 'line', Map(grid, 'name'), (0, 1), to_meters=False) == expected

def test_linear_ring():
    ls = LinearRing([(0.5, 0.5), (1.5, 0.5), (1.5, 1.5), (0.5, 1.5), (0.5, 0.5)])
    expected = {
        0: {
            'measure': 1,
            'geom': {'type': 'MultiLineString', 'coordinates': (((0.5, 0.5), (1.0, 0.5)), ((0.5, 1.0), (0.5, 0.5)))}},
        1: {
            'measure': 1,
            'geom': {'type': 'MultiLineString', 'coordinates': (((1.0, 1.5), (0.5, 1.5), (0.5, 1.0)),)}},
        2: {
            'measure': 1,
            'geom': {'type': 'MultiLineString', 'coordinates': (((1.0, 0.5), (1.5, 0.5), (1.5, 1.0)),)}},
        3: {
            'measure': 1,
            'geom': {'type': 'MultiLineString', 'coordinates': (((1.5, 1.0), (1.5, 1.5), (1.0, 1.5)),)}}
    }
    result = get_intersection(ls, 'line', Map(grid, 'name'), (0, 1, 2, 3), to_meters=False)
    assert result == expected

    expected = {
        0: {
            'measure': 1,
            'geom': {'type': 'MultiLineString', 'coordinates': (((0.5, 0.5), (1.0, 0.5)), ((0.5, 1.0), (0.5, 0.5)))}},
        1: {
            'measure': 1,
            'geom': {'type': 'MultiLineString', 'coordinates': (((1.0, 1.5), (0.5, 1.5), (0.5, 1.0)),)}}
    }
    assert get_intersection(ls, 'line', Map(grid, 'name'), (0, 1), to_meters=False) == expected

def test_line_geometry_collection():
    ls = GeometryCollection([LineString([(0.5, 0.5), (1.5, 0.5)])])
    expected = {
        0: {
            'geom': {'coordinates': (((0.5, 0.5), (1.0, 0.5)),), 'type': 'MultiLineString'},
            'measure': 0.5},
        2: {
            'geom': {'coordinates': (((1.0, 0.5), (1.5, 0.5)),), 'type': 'MultiLineString'},
            'measure': 0.5}
    }
    assert get_intersection(ls, 'line', Map(grid, 'name'), (0, 1, 2, 3), to_meters=False) == expected

    expected = {
        0: {
            'measure': 0.5,
            'geom': {'type': 'MultiLineString', 'coordinates': (((0.5, 0.5), (1.0, 0.5)),)}}
    }
    assert get_intersection(ls, 'line', Map(grid, 'name'), (0, 1), to_meters=False) == expected

def test_line_wrong_geometry():
    mp = Point((0.5, 1))
    assert get_intersection(mp, 'line', Map(grid, 'name'), (0, 1, 2, 3)) == {}

# Polygons

def test_polygon():
    pg = Polygon([(0.5, 0.5), (1.5, 0.5), (1.5, 1.5), (0.5, 1.5), (0.5, 0.5)])
    expected = {
        0: {
            'measure': 0.25,
            'geom': {'coordinates': [(((1.0, 0.5), (0.5, 0.5), (0.5, 1.0), (1.0, 1.0), (1.0, 0.5)),)], 'type': 'MultiPolygon'}},
        1: {
            'measure': 0.25,
            'geom': {'coordinates': [(((0.5, 1.0), (0.5, 1.5), (1.0, 1.5), (1.0, 1.0), (0.5, 1.0)),)], 'type': 'MultiPolygon'}},
        2: {
            'measure': 0.25,
            'geom': {'coordinates': [(((1.5, 1.0), (1.5, 0.5), (1.0, 0.5), (1.0, 1.0), (1.5, 1.0)),)], 'type': 'MultiPolygon'}},
        3: {
            'measure': 0.25,
            'geom': {'coordinates': [(((1.0, 1.5), (1.5, 1.5), (1.5, 1.0), (1.0, 1.0), (1.0, 1.5)),)], 'type': 'MultiPolygon'}}
    }
    assert get_intersection(pg, 'polygon', Map(grid, 'name'), (0, 1, 2, 3), to_meters=False) == expected

    expected = {
        0: {
            'measure': 0.25,
            'geom': {'coordinates': [(((1.0, 0.5), (0.5, 0.5), (0.5, 1.0), (1.0, 1.0), (1.0, 0.5)),)], 'type': 'MultiPolygon'}},
        1: {
            'measure': 0.25,
            'geom': {'coordinates': [(((0.5, 1.0), (0.5, 1.5), (1.0, 1.5), (1.0, 1.0), (0.5, 1.0)),)], 'type': 'MultiPolygon'}}
    }
    assert get_intersection(pg, 'polygon', Map(grid, 'name'), (0, 1), to_meters=False) == expected

def test_multi_polygon():
    pg = MultiPolygon([[
        [(0.5, 0.5), (1.5, 0.5), (1.5, 1.5), (0.5, 1.5), (0.5, 0.5)], []
    ]])
    expected = {
        0: {
            'geom': {'type': 'MultiPolygon', 'coordinates': [(((1.0, 0.5), (0.5, 0.5), (0.5, 1.0), (1.0, 1.0), (1.0, 0.5)),)]},
            'measure': 0.25},
        1: {
            'geom': {'type': 'MultiPolygon', 'coordinates': [(((0.5, 1.0), (0.5, 1.5), (1.0, 1.5), (1.0, 1.0), (0.5, 1.0)),)]},
            'measure': 0.25},
        2: {'geom': {'type': 'MultiPolygon', 'coordinates': [(((1.5, 1.0), (1.5, 0.5), (1.0, 0.5), (1.0, 1.0), (1.5, 1.0)),)]},
            'measure': 0.25},
        3: {
            'geom': {'type': 'MultiPolygon', 'coordinates': [(((1.0, 1.5), (1.5, 1.5), (1.5, 1.0), (1.0, 1.0), (1.0, 1.5)),)]},
            'measure': 0.25}
    }
    assert get_intersection(pg, 'polygon', Map(grid, 'name'), (0, 1, 2, 3), to_meters=False) == expected

    expected = {
        0: {
            'measure': 0.25,
            'geom': {'coordinates': [(((1.0, 0.5), (0.5, 0.5), (0.5, 1.0), (1.0, 1.0), (1.0, 0.5)),)], 'type': 'MultiPolygon'}},
        1: {
            'measure': 0.25,
            'geom': {'coordinates': [(((0.5, 1.0), (0.5, 1.5), (1.0, 1.5), (1.0, 1.0), (0.5, 1.0)),)], 'type': 'MultiPolygon'}}
    }
    assert get_intersection(pg, 'polygon', Map(grid, 'name'), (0, 1), to_meters=False) == expected

def test_polygon_geometry_collection():
    pg = GeometryCollection([Polygon([(0.5, 0.5), (1.5, 0.5), (1.5, 1.5), (0.5, 1.5), (0.5, 0.5)])])
    expected = {
        0: {
            'geom': {'coordinates': [(((1.0, 0.5), (0.5, 0.5), (0.5, 1.0), (1.0, 1.0), (1.0, 0.5)),)], 'type': 'MultiPolygon'},
            'measure': 0.25},
        1: {
            'geom': {'coordinates': [(((0.5, 1.0), (0.5, 1.5), (1.0, 1.5), (1.0, 1.0), (0.5, 1.0)),)], 'type': 'MultiPolygon'},
            'measure': 0.25},
        2: {
            'geom': {'coordinates': [(((1.5, 1.0), (1.5, 0.5), (1.0, 0.5), (1.0, 1.0), (1.5, 1.0)),)], 'type': 'MultiPolygon'},
            'measure': 0.25},
        3: {
            'geom': {'coordinates': [(((1.0, 1.5), (1.5, 1.5), (1.5, 1.0), (1.0, 1.0), (1.0, 1.5)),)], 'type': 'MultiPolygon'},
            'measure': 0.25}
    }
    assert get_intersection(pg, 'polygon', Map(grid, 'name'), (0, 1, 2, 3), to_meters=False) == expected

    expected = {
        0: {
            'geom': {'type': 'MultiPolygon', 'coordinates': [(((1.0, 0.5), (0.5, 0.5), (0.5, 1.0), (1.0, 1.0), (1.0, 0.5)),)]},
            'measure': 0.25},
        1: {'geom': {'type': 'MultiPolygon', 'coordinates': [(((0.5, 1.0), (0.5, 1.5), (1.0, 1.5), (1.0, 1.0), (0.5, 1.0)),)]},
            'measure': 0.25}
    }
    assert get_intersection(pg, 'polygon', Map(grid, 'name'), (0, 1), to_meters=False) == expected

def test_polygon_wrong_geometry():
    mp = Point((0.5, 1))
    assert get_intersection(mp, 'polygon', Map(grid, 'name'), (0, 1, 2, 3)) == {}

# Clean

def test_clean():
    p = Polygon([(0,0), (0,3), (3,3), (3,0), (2,0),
                 (2,2), (1,2), (1,1), (2,1), (2,0), (0,0)])
    pp = clean(p)
    assert not p.is_valid
    assert pp.is_valid

# Get measure

def test_get_measure_point():
    mp = MultiPoint([(0.5, 0.5), (0.5, 1), (1, 1), (1.5, 1.5)])
    assert get_measure(mp) == 4
    assert get_measure(mp, 'point') == 4

    mp = Point((0.5, 1))
    assert get_measure(mp) == 1
    assert get_measure(mp, 'point') == 1

def test_get_measure_point_not_point_geom():
    geom = Polygon([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)])
    with pytest.raises(ValueError):
        get_measure(geom, 'point')

def test_get_measure_line():
    geom = LineString([(0, 0), (0, 1)])
    assert get_measure(geom) == 1
    assert get_measure(geom, 'line') == 1

    geom = MultiLineString([
        [(0.5, 0.5), (1.5, 0.5)],
        [(1, 1), (1, 11)]
    ])
    assert get_measure(geom) == 11
    assert get_measure(geom, 'line') == 11

    geom = LinearRing([(0.5, 0.5), (1.5, 0.5), (1.5, 1.5), (0.5, 1.5), (0.5, 0.5)])
    assert get_measure(geom) == 4
    assert get_measure(geom, 'line') == 4

def test_get_measure_polygon():
    geom = Polygon([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)])
    assert get_measure(geom) == 1
    assert get_measure(geom, 'polygon') == 1

    geom = MultiPolygon([[
        [(0, 0), (0, 2), (2, 2), (2, 0), (0 ,0)],
        [
            [(0.5, 0.5), (0.5, 1.5), (1.5, 1.5), (1.5, 0.5), (0.5, 0.5)]
        ]
    ], [
        [(10, 10), (10, 17), (11, 17), (11, 10), (10, 10)],
        []
    ]])
    assert get_measure(geom) == 10
    assert get_measure(geom, 'polygon') == 10

def test_get_measure_wrong_type():
    mp = MultiPoint([(0.5, 0.5), (0.5, 1), (1, 1), (1.5, 1.5)])
    gc = GeometryCollection([mp])
    with pytest.raises(ValueError):
        get_measure(gc)

    geom = Polygon([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)])
    get_measure(geom)
    with pytest.raises(ValueError):
        get_measure(geom, 'foo')

# Remaining calculations

def test_get_remaining_wrong_type():
    mp = MultiPoint([(0.5, 0.5), (0.5, 1), (1, 1), (1.5, 1.5)])
    gc = GeometryCollection([mp])
    with pytest.raises(ValueError):
        get_remaining(gc, [])

def test_remaining_incompatible_types():
    geom = Polygon([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)])
    other = LineString([(0, .5), (0, 1)])
    with pytest.raises(IncompatibleTypes):
        get_remaining(geom, [other])

# Polygons

def test_remaining_polygons():
    geom = Polygon([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)])
    half = Polygon([(0, 0), (0, .5), (1, .5), (1, 0), (0, 0)])
    assert get_remaining(geom, [half], False) == 0.5

    second = Polygon([(0, .5), (0, 1), (1, 1), (1, .5), (0, .5)])
    assert get_remaining(geom, [half, second], False) == 0

def test_remaining_polygons_projection():
    area = 1/2 * (4e7 / 360) ** 2
    geom = Polygon([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)])
    half = Polygon([(0, 0), (0, .5), (1, .5), (1, 0), (0, 0)])
    assert np.isclose(get_remaining(geom, [half]), area, 1e-2)

def test_remaining_polygons_no_geoms():
    geom = Polygon([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)])
    assert get_remaining(geom, [], False) == 1

# Lines

def test_remaining_lines():
    geom = LineString([(0, 0), (0, 1), (1, 1)])
    half = LineString([(0, 0), (0, 1)])
    assert get_remaining(geom, [half], False) == 1

    second = LineString([(0, 1), (1, 1)])
    assert get_remaining(geom, [half, second], False) == 0

def test_remaining_lines_overlap():
    geom = LineString([(0, 0), (0, 1), (1, 1)])
    half = LineString([(0, 0), (0, 1)])
    quarter = LineString([(0, 0.5), (0, 1)])
    result = get_remaining(geom, [half, quarter], False)
    assert result == (2 - 1) * (1.5/1)

def test_remaining_lines_projection():
    length = 4e7 / 360
    geom = LineString([(0, 0), (0, 1), (1, 1)])
    half = LineString([(0, 0), (0, 1)])
    assert np.isclose(get_remaining(geom, [half]), 1e5, 1e-2)

def test_remaining_lines_no_geoms():
    geom = LineString([(0, 0), (0, 1), (1, 1)])
    assert get_remaining(geom, [], False) == 2

# Points

def test_remaining_points():
    geom = MultiPoint([(0, 0), (0, 1)])
    half = Point((0, 0))
    assert get_remaining(geom, [half], False) == 1

def test_remaining_points_overlap():
    geom = MultiPoint([(0, 0), (0, 1)])
    first = Point((0, 0))
    second = Point((0, 0))
    result = get_remaining(geom, [first, second], False)
    assert result == (2 - 1) * (2 / 1)

def test_remaining_points_projection():
    geom = MultiPoint([(0, 0), (0, 1)])
    half = Point((0, 0))
    assert get_remaining(geom, [half]) == 1

def test_remaining_points_no_geoms():
    geom = MultiPoint([(0, 0), (0, 1)])
    assert get_remaining(geom, [], False) == 2



from pandarus.filesystem import sha256, json_exporter, get_appdirs_path, json_importer
import tempfile
import os

dirpath = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))


def test_hashing():
    assert sha256(os.path.join(dirpath, "testfile.hash")) == \
        'd2adeda32326a6576b73f9f387d75798d5bd6f0b4d385d36684fdb7d205a0ab0'


def test_json_exporting():
    with tempfile.TemporaryDirectory() as dirpath:
        new_fp = os.path.join(dirpath, 'testfile')
        fp = json_exporter({'d': [1,2,3], 'e': {'foo': 'bar'}}, new_fp, False)
        assert fp
        assert not fp.endswith(".bz2")
        assert os.path.isfile(fp)

        fp = json_exporter([1,2,3], new_fp, True)
        assert fp
        assert fp.endswith(".bz2")
        assert os.path.isfile(fp)

def test_json_importing():
    fp = os.path.join(dirpath, "json_test.json")
    assert json_importer(fp) == {'d': [1,2,3], 'e': {'foo': 'bar'}}

    fp = os.path.join(dirpath, "json_test.json.bz2")
    assert json_importer(fp) == {'d': [1,2,3], 'e': {'foo': 'bar'}}

def test_json_roundtrip():
    data = {'d': [1,2,3], 'e': {'foo': 'bar'}}

    with tempfile.TemporaryDirectory() as dirpath:
        new_fp = os.path.join(dirpath, 'testfile')

        fp = json_exporter(data, new_fp, False)
        assert json_importer(fp) == data

        fp = json_exporter(data, new_fp)
        assert json_importer(fp) == data

def test_appdirs_path():
    dp = get_appdirs_path("test-dir")
    assert os.path.exists(dp)
    assert os.path.isdir(dp)
    assert "test-dir" in dp
    assert "pandarus" in dp

    os.rmdir(dp)
    assert not os.path.exists(dp)
    dp = get_appdirs_path("test-dir")
    assert os.path.exists(dp)
    os.rmdir(dp)



from pandarus.projection import MOLLWEIDE, WGS84, project, wgs84
from shapely.geometry import (
    GeometryCollection,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
)


def test_projection():
    given = GeometryCollection([
        MultiPoint([(1, 2), (3, 4), (5, 6), (7, 8)]),
        MultiLineString([[(20, 30), (40, 50), (60, 70)]]),
        MultiPolygon([[[(1.1, 1.2), (1.3, 1.4), (1.5, 1.6), (1.7, 1.8), (1.1, 1.2)], []]]),
    ])
    result = project(given, WGS84, MOLLWEIDE).wkt
    assert 'MULTIPOINT' in result
    assert 'GEOMETRYCOLLECTION' in result

    result = project(given).wkt
    assert 'MULTIPOINT' in result
    assert 'GEOMETRYCOLLECTION' in result


    expected = 'GEOMETRYCOLLECTION (MULTIPOINT (1 2, 3 4, 5 6, 7 8), MULTILINESTRING ((20 30, 40 50, 60 70)), MULTIPOLYGON (((1.1 1.2, 1.3 1.4, 1.5 1.6, 1.7 1.8, 1.1 1.2))))'
    assert project(given, WGS84, WGS84).wkt == expected


def test_wgs84():
    """Fix no CRS or fiona giving abbreviated wgs84 definition."""
    assert wgs84(None) == WGS84
    assert wgs84("+no_defs") == WGS84
    assert wgs84(1) == 1



from pandarus import Map
from pandarus.intersections import (
    chunker,
    get_jobs,
    intersection_worker,
    intersection_dispatcher,
    logger_init,
    worker_init,
)
import os
import numpy as np
import pytest
import tempfile

dirpath = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))
grid = os.path.join(dirpath, "grid.geojson")
square = os.path.join(dirpath, "square.geojson")
point = os.path.join(dirpath, "point.geojson")
gc = os.path.join(dirpath, "gc.geojson")


def test_chunker():
    numbers = list(range(10))
    expected = [
        [0, 1, 2, 3],
        [4, 5, 6, 7],
        [8, 9]
    ]
    assert list(chunker(numbers, 4)) == expected

def test_worker_init():
    with tempfile.TemporaryDirectory() as dirpath:
        ql, lq = logger_init(dirpath)
        worker_init(lq)
        assert os.listdir(dirpath)

def test_get_jobs():
    assert get_jobs(10) == (20, 1)
    assert get_jobs(100) == (20, 5)
    assert get_jobs(110) == (20, 6)
    assert get_jobs(10000) == (50, 200)

def test_intersection_worker_indices():
    area = 1/4 * (4e7 / 360) ** 2
    result = intersection_worker(grid, [0], square)
    assert result.keys() == {(0, 0)}
    key, value = list(result.keys())[0], list(result.values())[0]
    expected = 'MULTIPOLYGON (((0.5 1, 1 1, 1 0.5, 0.5 0.5, 0.5 1)))'
    assert value['geom'].wkt == expected
    assert np.isclose(value['measure'], area, rtol=1e-2)

def test_intersection_worker_no_indices():
    result = intersection_worker(grid, None, square)
    assert len(result) == 4

def test_intersection_worker_wrong_from_type():
    with pytest.raises(ValueError):
        intersection_worker(gc, None, square)

def test_intersection_worker_wrong_to_type():
    with pytest.raises(ValueError):
        intersection_worker(grid, [0], point)

def test_intersection_dispatcher_zero_cpus():
    result = intersection_dispatcher(grid, square)
    assert len(result) == 4

def test_intersection_dispatcher_indices():
    with tempfile.TemporaryDirectory() as dirpath:
        result = intersection_dispatcher(grid, square, [0, 1], 1, dirpath)
        assert len(result) == 2



from pandarus import intersect
from pandarus.conversion import round_to_x_significant_digits
from math import sqrt
import fiona
import json
import numpy as np
import os
import tempfile

dirpath = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))
grid = os.path.join(dirpath, "grid.geojson")
square = os.path.join(dirpath, "square.geojson")
range_raster = os.path.join(dirpath, "range.tif")
dem = os.path.join(dirpath, "DEM.tif")
outside = os.path.join(dirpath, "outside.geojson")
remain_result = os.path.join(dirpath, "remaining.geojson")
pgrid = os.path.join(dirpath, "grid-3410.geojson")
igrid = os.path.join(dirpath, "grid-ints.geojson")
psquare = os.path.join(dirpath, "square-3857.geojson")
plines =  os.path.join(dirpath, "lines-25000.geojson")
lines =  os.path.join(dirpath, "lines.geojson")
ppoints =  os.path.join(dirpath, "points-32631.geojson")
points =  os.path.join(dirpath, "points.geojson")


def test_intersection_polygon():
    area = 1/4 * (4e7 / 360) ** 2

    with tempfile.TemporaryDirectory() as dirpath:
        vector_fp, data_fp = intersect(
            outside,
            'name',
            grid,
            'name',
            dirpath=dirpath,
            compress=False,
            log_dir=dirpath
        )
        data = json.load(open(data_fp))

        assert len(data['data']) == 2
        for x, y, z in data['data']:
            assert x == 'by-myself'
            assert y in ('grid cell 1', 'grid cell 3')
            assert np.isclose(z, area, rtol=1e-2)

        assert data['metadata'].keys() == {'first', 'second', 'when'}
        assert data['metadata']['first'].keys() == {'field', 'filename', 'path', 'sha256'}
        assert data['metadata']['second'].keys() == {'field', 'filename', 'path', 'sha256'}

        with fiona.open(vector_fp) as src:
            meta = src.meta

            assert meta['driver'] == 'GeoJSON'
            assert meta['schema'] == {
                'geometry': 'MultiPolygon',
                'properties': dict([
                    ('measure', 'float'),
                    ('from_label', 'str'),
                    ('id', 'int'),
                    ('to_label', 'str')
                ])
            }
            assert meta['crs'] == {'init': 'epsg:4326'}

            coords = [
                [[[(0.5, 1.5), (0.5, 2.0), (1.0, 2.0), (1.0, 1.5), (0.5, 1.5)]]],
                [[[(1.5, 2.0), (1.5, 1.5), (1.0, 1.5), (1.0, 2.0), (1.5, 2.0)]]]
            ]

            for feature in src:
                assert feature['geometry']['coordinates'] in coords
                assert feature['geometry']['type'] == 'MultiPolygon'
                assert feature['properties'].keys() == {'measure', 'from_label', 'to_label', 'id'}
                assert np.isclose(feature['properties']['measure'], area, rtol=1e-2)

def test_intersection_polygon_integer_indices():
    area = 1/4 * (4e7 / 360) ** 2

    with tempfile.TemporaryDirectory() as dirpath:
        vector_fp, data_fp = intersect(
            outside,
            'name',
            igrid,
            'name',
            dirpath=dirpath,
            compress=False,
            log_dir=dirpath
        )
        data = json.load(open(data_fp))

        assert len(data['data']) == 2
        for x, y, z in data['data']:
            assert x == 'by-myself'
            assert y in (1, 3)
            assert np.isclose(z, area, rtol=1e-2)

        assert data['metadata'].keys() == {'first', 'second', 'when'}
        assert data['metadata']['first'].keys() == {'field', 'filename', 'path', 'sha256'}
        assert data['metadata']['second'].keys() == {'field', 'filename', 'path', 'sha256'}

        with fiona.open(vector_fp) as src:
            meta = src.meta

            assert meta['driver'] == 'GeoJSON'
            assert meta['schema'] == {
                'geometry': 'MultiPolygon',
                'properties': dict([
                    ('measure', 'float'),
                    ('from_label', 'str'),
                    ('id', 'int'),
                    ('to_label', 'int')
                ])
            }
            assert meta['crs'] == {'init': 'epsg:4326'}

            coords = [
                [[[(0.5, 1.5), (0.5, 2.0), (1.0, 2.0), (1.0, 1.5), (0.5, 1.5)]]],
                [[[(1.5, 2.0), (1.5, 1.5), (1.0, 1.5), (1.0, 2.0), (1.5, 2.0)]]]
            ]

            for feature in src:
                assert feature['geometry']['coordinates'] in coords
                assert feature['geometry']['type'] == 'MultiPolygon'
                assert feature['properties'].keys() == {'measure', 'from_label', 'to_label', 'id'}
                assert np.isclose(feature['properties']['measure'], area, rtol=1e-2)

def test_intersection_polygon_projection():
    area = 1/4 * (4e7 / 360) ** 2

    with tempfile.TemporaryDirectory() as dirpath:
        vector_fp, data_fp = intersect(
            pgrid,
            'name',
            psquare,
            'name',
            dirpath=dirpath,
            compress=False,
            log_dir=dirpath
        )
        data = json.load(open(data_fp))

        assert len(data['data']) == 4
        for x, y, z in data['data']:
            assert x in ['grid cell {}'.format(x) for x in range(4)]
            assert y == 'single'
            assert np.isclose(z, area, rtol=1e-2)

        assert data['metadata'].keys() == {'first', 'second', 'when'}
        assert data['metadata']['first'].keys() == {'field', 'filename', 'path', 'sha256'}
        assert data['metadata']['second'].keys() == {'field', 'filename', 'path', 'sha256'}

        with fiona.open(vector_fp) as src:
            meta = src.meta

            assert meta['driver'] == 'GeoJSON'
            assert meta['schema'] == {
                'geometry': 'MultiPolygon',
                'properties': dict([
                    ('measure', 'float'),
                    ('from_label', 'str'),
                    ('id', 'int'),
                    ('to_label', 'str')
                ])
            }
            assert meta['crs'] == {'init': 'epsg:4326'}

            coords = [
                [[[(0.5, 1.0), (1.0, 1.0), (1.0, 0.5), (0.5, 0.5), (0.5, 1.0)]]],
                [[[(1.0, 1.5), (1.0, 1.0), (0.5, 1.0), (0.5, 1.5), (1.0, 1.5)]]],
                [[[(1.0, 0.5), (1.0, 1.0), (1.5, 1.0), (1.5, 0.5), (1.0, 0.5)]]],
                [[[(1.0, 1.0), (1.0, 1.5), (1.5, 1.5), (1.5, 1.0), (1.0, 1.0)]]]
            ]

            for feature in src:
                print(feature['geometry']['coordinates'])
                assert feature['geometry']['coordinates'] in coords
                assert feature['geometry']['type'] == 'MultiPolygon'
                assert feature['properties'].keys() == {'measure', 'from_label', 'to_label', 'id'}
                assert np.isclose(feature['properties']['measure'], area, rtol=1e-2)

def test_intersection_line():
    one_degree = 4e7 / 360

    with tempfile.TemporaryDirectory() as dirpath:
        vector_fp, data_fp = intersect(
            lines,
            'name',
            grid,
            'name',
            dirpath=dirpath,
            compress=False,
            log_dir=dirpath
        )
        data = json.load(open(data_fp))
        data_dct = {(x, y): z for x, y, z in data['data']}

        assert len(data['data']) == 4
        assert np.isclose(data_dct[('A', 'grid cell 0')], 62000, rtol=1e-2)
        assert np.isclose(data_dct[('A', 'grid cell 1')], one_degree, rtol=1e-2)
        assert np.isclose(data_dct[('A', 'grid cell 3')], 50000, rtol=1e-2)
        assert np.isclose(data_dct[('B', 'grid cell 2')], sqrt(2) * one_degree / 2, rtol=2e-2)

        assert data['metadata'].keys() == {'first', 'second', 'when'}
        assert data['metadata']['first'].keys() == {'field', 'filename', 'path', 'sha256'}
        assert data['metadata']['second'].keys() == {'field', 'filename', 'path', 'sha256'}

        with fiona.open(vector_fp) as src:
            meta = src.meta

            assert meta['driver'] == 'GeoJSON'
            assert meta['schema'] == {
                'geometry': 'MultiLineString',
                'properties': dict([
                    ('measure', 'float'),
                    ('from_label', 'str'),
                    ('id', 'int'),
                    ('to_label', 'str')
                ])
            }
            assert meta['crs'] == {'init': 'epsg:4326'}

            coords = [
                [[(1.0, 1.5), (1.5, 1.5)]],
                [[(0.5, 1.0), (0.5, 1.5), (1.0, 1.5)]],
                [[(0.5, 0.5), (0.5, 1.0)]],
                [[(1.0, 1.0), (1.5, 0.5)]]
            ]

            for feature in src:
                assert feature['geometry']['coordinates'] in coords
                assert feature['geometry']['type'] == 'MultiLineString'
                assert feature['properties'].keys() == {'measure', 'from_label', 'to_label', 'id'}

def test_intersection_line_projection():
    one_degree = 4e7 / 360

    with tempfile.TemporaryDirectory() as dirpath:
        vector_fp, data_fp = intersect(
            plines,
            'name',
            pgrid,
            'name',
            dirpath=dirpath,
            compress=False,
            log_dir=dirpath
        )
        data = json.load(open(data_fp))
        data_dct = {(x, y): z for x, y, z in data['data']}

        assert len(data['data']) == 5
        assert np.isclose(data_dct[('A', 'grid cell 0')], 62000, rtol=1e-2)
        assert np.isclose(data_dct[('A', 'grid cell 1')], one_degree, rtol=1e-2)
        assert np.isclose(data_dct[('A', 'grid cell 3')], 50000, rtol=1e-2)
        assert np.isclose(data_dct[('B', 'grid cell 2')], sqrt(2) * one_degree / 2, rtol=2e-2)
        assert data_dct[('B', 'grid cell 3')] < 1e-3

        assert data['metadata'].keys() == {'first', 'second', 'when'}
        assert data['metadata']['first'].keys() == {'field', 'filename', 'path', 'sha256'}
        assert data['metadata']['second'].keys() == {'field', 'filename', 'path', 'sha256'}

        with fiona.open(vector_fp) as src:
            meta = src.meta

            assert meta['driver'] == 'GeoJSON'
            assert meta['schema'] == {
                'geometry': 'MultiLineString',
                'properties': dict([
                    ('measure', 'float'),
                    ('from_label', 'str'),
                    ('id', 'int'),
                    ('to_label', 'str')
                ])
            }
            assert meta['crs'] == {'init': 'epsg:4326'}

            coords = [np.array(x) for x in [
                [[[1., 1.], [1., 1.]]],
                [[(1.0, 1.5), (1.5, 1.5)]],
                [[(0.5, 1.0), (0.5, 1.5), (1.0, 1.5)]],
                [[(0.5, 0.5), (0.5, 1.0)]],
                [[(1.0, 1.0), (1.5, 0.5)]]
            ]]

            arrays = [
                round_to_x_significant_digits(np.array(x['geometry']['coordinates']))
                for x in src
            ]

            for array in arrays:
                print(array)
                assert any(np.allclose(array, obj)
                    for obj in coords
                    if array.shape == obj.shape)

            for feature in src:
                assert feature['geometry']['type'] == 'MultiLineString'
                assert feature['properties'].keys() == {'measure', 'from_label', 'to_label', 'id'}

def test_intersection_point():
    with tempfile.TemporaryDirectory() as dirpath:
        vector_fp, data_fp = intersect(
            points,
            'name',
            grid,
            'name',
            dirpath=dirpath,
            compress=False,
            log_dir=dirpath
        )
        data = json.load(open(data_fp))
        data_dct = {(x, y): z for x, y, z in data['data']}

        assert sorted(data['data']) == sorted([['point 1', 'grid cell 0', 1.0], ['point 2', 'grid cell 3', 1.0]])

        assert data['metadata'].keys() == {'first', 'second', 'when'}
        assert data['metadata']['first'].keys() == {'field', 'filename', 'path', 'sha256'}
        assert data['metadata']['second'].keys() == {'field', 'filename', 'path', 'sha256'}

        with fiona.open(vector_fp) as src:
            meta = src.meta

            assert meta['driver'] == 'GeoJSON'
            assert meta['schema'] == {
                'geometry': 'MultiPoint',
                'properties': dict([
                    ('measure', 'float'),
                    ('from_label', 'str'),
                    ('id', 'int'),
                    ('to_label', 'str')
                ])
            }
            assert meta['crs'] == {'init': 'epsg:4326'}

            coords = [[(0.5, 0.5)], [(1.5, 1.5)]]

            for feature in src:
                assert feature['geometry']['coordinates'] in coords
                assert feature['geometry']['type'] == 'MultiPoint'
                assert feature['properties'].keys() == {'measure', 'from_label', 'to_label', 'id'}

def test_intersection_point_projection():
    with tempfile.TemporaryDirectory() as dirpath:
        vector_fp, data_fp = intersect(
            ppoints,
            'name',
            pgrid,
            'name',
            dirpath=dirpath,
            compress=False,
            log_dir=dirpath
        )
        data = json.load(open(data_fp))
        data_dct = {(x, y): z for x, y, z in data['data']}

        assert len(data['data']) == 2
        assert data_dct[('point 1', 'grid cell 0')] == 1
        assert data_dct[('point 2', 'grid cell 3')] == 1
        assert data['metadata'].keys() == {'first', 'second', 'when'}
        assert data['metadata']['first'].keys() == {'field', 'filename', 'path', 'sha256'}
        assert data['metadata']['second'].keys() == {'field', 'filename', 'path', 'sha256'}

        with fiona.open(vector_fp) as src:
            meta = src.meta

            assert meta['driver'] == 'GeoJSON'
            assert meta['schema'] == {
                'geometry': 'MultiPoint',
                'properties': dict([
                    ('measure', 'float'),
                    ('from_label', 'str'),
                    ('id', 'int'),
                    ('to_label', 'str')
                ])
            }
            assert meta['crs'] == {'init': 'epsg:4326'}

            coords = [[(0.5, 0.5)], [(1.5, 1.5)]]

            for feature in src:
                print(feature)
                assert feature['geometry']['coordinates'] in coords
                assert feature['geometry']['type'] == 'MultiPoint'
                assert feature['properties'].keys() == {'measure', 'from_label', 'to_label', 'id'}



from pandarus.conversion import *
from affine import Affine
from rasterio.crs import CRS
import fiona
import numpy as np
import os
import pytest
import rasterio
import tempfile

dirpath = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))
grid = os.path.join(dirpath, "grid.geojson")
sixty_four = os.path.join(dirpath, "test_raster_cfs.tif")
cfs = os.path.join(dirpath, "raster_cfs_32bit.tif")
dem = os.path.join(dirpath, "DEM.tif")
invalid = os.path.join(dirpath, "invalid.txt")


def create_raster(name, array, dirpath, nodata=-1, **kwargs):
    profile = {
        'count': 1,
        'nodata': nodata,
        'dtype': 'float32',
        'width': array.shape[1],
        'height': array.shape[0],
        'affine': Affine(0.1, 0, 10, 0, -0.1, 10),
        'driver': 'GTiff',
        'compress': 'lzw',
        'crs': CRS.from_epsg(4326)
    }
    profile.update(kwargs)
    fp = os.path.join(dirpath, name)

    with rasterio.Env():
        with rasterio.open(fp, 'w', **profile) as dst:
            dst.write(array, 1)

    return fp

def test_create_raster():
    with tempfile.TemporaryDirectory() as dirpath:
        fp = create_raster(
            'foo.tif',
            np.random.random(size=(10,10)).astype(np.float32),
            dirpath
        )
        assert os.path.isfile(fp)
        assert check_type(fp) == 'raster'
        with rasterio.open(fp) as f:
            assert f.read(1).shape == (10, 10)
            assert f.read(1).dtype == np.float32

def test_check_type():
    assert check_type(grid) == 'vector'
    assert check_type(cfs) == 'raster'
    with pytest.raises(ValueError):
        check_type(invalid)

def test_convert_to_vector():
    with pytest.raises(AssertionError):
        convert_to_vector(cfs, band='1')

    with tempfile.TemporaryDirectory() as dirpath:
        out = convert_to_vector(cfs, dirpath)
        assert check_type(out) == 'vector'

        # Second time should be a no-op
        out = convert_to_vector(cfs, dirpath)
        assert check_type(out) == 'vector'

        with fiona.open(out) as src:
            meta = src.meta

            assert meta['crs'] == {'init': 'epsg:4326'}
            assert meta['schema']['properties'].keys() == {'filename', 'id', 'val'}
            assert meta['schema']['geometry'] in ('Polygon', 'MultiPolygon')

    out = convert_to_vector(cfs)
    assert check_type(out) == 'vector'
    os.remove(out)

def test_rounding():
    given = np.array([3.14159358979, 2.718281828459045235360, 325796139])
    expected = (3.142, 2.718, 3.258e8)
    for x, y in zip(round_to_x_significant_digits(given, 4), expected):
        assert x == y

def test_round_raster():
    with tempfile.TemporaryDirectory() as dirpath:
        out = os.path.join(dirpath, "test.tif")

        assert round_raster(cfs, out) == out

        with rasterio.open(out) as src:
            array = src.read(1)
            profile = src.profile

    assert profile['driver'] == 'GTiff'
    assert profile['compress'] == 'lzw'
    assert profile['count'] == 1
    assert profile['height'] == 16
    assert np.unique(array[:3, :3]).shape == (1,)
    assert np.isclose(array[0, 0], 1.47e-7)

def test_clean_raster():
    out = clean_raster(sixty_four)

    with rasterio.open(out) as src:
        array = src.read(1)
        profile = src.profile

    assert array.dtype == np.float32
    assert profile['driver'] == 'GTiff'
    assert profile['compress'] == 'lzw'
    assert profile['count'] == 1
    assert profile['nodata'] == -1
    assert not profile['tiled']
    assert 'blockysize' not in profile
    assert 'blockxsize' not in profile
    os.remove(out)

def test_clean_raster_null_nodata():
    with tempfile.TemporaryDirectory() as dirpath:
        out = os.path.join(dirpath, "test.tif")
        result = clean_raster(dem, out)

        with rasterio.open(out) as src:
            profile = src.profile

    assert profile['nodata'] is None

def test_clean_raster_filepath():
    with tempfile.TemporaryDirectory() as dirpath:
        out = os.path.join(dirpath, "test.tif")
        result = clean_raster(sixty_four, out)
        assert result == out

def test_clean_raster_64bit():
    array = np.array([[0, -1, 1e100]])

    with tempfile.TemporaryDirectory() as dirpath:
        fp = create_raster('foo.tif', array, dirpath, dtype='float64', nodata=42)

        with rasterio.open(fp) as f:
            assert f.profile['nodata'] == 42
            assert f.read(1).dtype == np.float64

        out_fp = os.path.join(dirpath, 'clean.tif')
        out = clean_raster(fp, out_fp)

        with rasterio.open(fp) as f:
            assert f.read(1).dtype == np.float64

def test_clean_raster_out_of_bounds():
    array = np.array([[0, 1.5, 42, -1e50]])
    with tempfile.TemporaryDirectory() as dirpath:
        fp = create_raster('foo.tif', array, dirpath, dtype='float64', nodata=None)
        out = clean_raster(fp)
        assert out is None

def test_clean_raster_dont_change():
    array = np.array([[0, 1.5, 42, -1e7]])
    with tempfile.TemporaryDirectory() as dirpath:
        fp = create_raster('foo.tif', array, dirpath, dtype='float64', nodata=-1e7)
        out = clean_raster(fp)
        with rasterio.open(out) as f:
            assert f.profile['nodata'] == -1e7
    os.remove(out)

def test_clean_raster_nodata():
    array = np.array([[0, 1.5, 42, -1e50]])
    with tempfile.TemporaryDirectory() as dirpath:
        fp = create_raster('foo.tif', array, dirpath, dtype='float64', nodata=-1e50)
        out = clean_raster(fp)
        with rasterio.open(out) as f:
            assert f.profile['nodata'] == -1
    os.remove(out)

    array = np.array([[0, -1., -99., 6/7]])
    with tempfile.TemporaryDirectory() as dirpath:
        fp = create_raster('foo.tif', array, dirpath, dtype='float64', nodata=-1e50)
        out = clean_raster(fp)
    with rasterio.open(out) as f:
        assert f.profile['nodata'] == -999
    os.remove(out)

    array = np.array([[0, -1., -99., -999., -9999.]])
    with tempfile.TemporaryDirectory() as dirpath:
        fp = create_raster('foo.tif', array, dirpath, dtype='float64', nodata=-1e50)
        assert clean_raster(fp) is None

def test_clean_raster_try_given_nodata():
    array = np.array([[0, -1., -99., -999., -9999]])
    with tempfile.TemporaryDirectory() as dirpath:
        fp = create_raster('foo.tif', array, dirpath, dtype='float64', nodata=-1e50)
        out = clean_raster(fp, nodata=42)
        with rasterio.open(out) as f:
            assert f.profile['nodata'] == 42
    os.remove(out)



# -*- coding: utf-8 -*-
from affine import Affine
from fiona import crs as fiona_crs
from rasterio.crs import CRS
import fiona
import itertools
import json
import numpy as np
import os
import rasterio


dirpath = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))

wgs84 = fiona_crs.from_string("+datum=WGS84 +ellps=WGS84 +no_defs +proj=longlat")

"""Valid geometry types:

- Point
- LineString
- Polygon
- MultiPoint
- MultiLineString
- MultiPolygon

"""

def create_raster(fp, array, nodata=-1, **kwargs):
    profile = {
        'count': 1,
        'nodata': nodata,
        'dtype': 'float32',
        'width': array.shape[1],
        'height': array.shape[0],
        'affine': Affine(0.4, 0, 0, 0, -0.2, 2),
        'driver': 'GTiff',
        'compress': 'lzw',
        'crs': CRS.from_epsg(4326)
    }
    profile.update(kwargs)

    with rasterio.Env():
        with rasterio.open(fp, 'w', **profile) as dst:
            dst.write(array, 1)

    return fp


def create_schema(geometry='Polygon'):
    return {'geometry': geometry, 'properties': {'name': 'str'}}


def create_record(name, coords, geometry='Polygon'):
    return {
        'geometry': {'coordinates': coords, 'type': geometry},
        'properties': {'name': name}
    }


def create_test_file(filepath, records, schema=None, driver='GeoJSON', crs=wgs84):
    if os.path.exists(filepath):
        os.remove(filepath)
    if schema is None:
        schema = create_schema()
    with fiona.Env():
        with fiona.open(filepath, 'w', driver=driver, crs=crs, schema=schema) as outfile:
            for record in records:
                outfile.write(record)


def create_box(x, y, width, height):
    return [[
        (x, y),
        (x, y + height),
        (x + width, y + height),
        (x + width, y),
        (x, y)
    ]]


def create_grid(start_x=0.0, start_y=0.0, cols=4, rows=4, width=1.0,
        height=1.0):
    data = []
    for x_increment in range(cols):
        for y_increment in range(rows):
            data.append(create_box(
                start_x + x_increment * width,
                start_y + y_increment * height,
                width,
                height,
                ))
    return data


def create_test_datasets():
    # Create 4x4 grid
    x, y, cols, rows = 0, 0, 2, 2
    names = ["grid cell {}".format(i * 2 + j)
        for i in range(cols)
        for j in range(rows)
    ]
    cells = create_grid(x, y, cols, rows)
    create_test_file(
        os.path.join(dirpath, "grid.geojson"),
        [create_record(name, coords) for name, coords in zip(names, cells)]
    )

    names = ["grid cell {}".format(i * 2 + j)
        for i in range(cols)
        for j in range(rows)
    ]
    cells = create_grid(0, 7, cols, rows)
    create_test_file(
        os.path.join(dirpath, "big-grid.geojson"),
        [create_record(name, coords) for name, coords in zip(names, cells)]
    )

    # Duplicate names
    names = itertools.repeat("foo")
    cells = create_grid(x, y, cols, rows)
    create_test_file(
        os.path.join(dirpath, "duplicates.geojson"),
        [create_record(name, coords) for name, coords in zip(names, cells)]
    )

    # Create one square
    create_test_file(
        os.path.join(dirpath, "square.geojson"),
        [create_record("single", create_box(0.5, 0.5, 1, 1))]
    )

    create_test_file(
        os.path.join(dirpath, "outside.geojson"),
        [create_record("by-myself", create_box(0.5, 1.5, 1, 1))]
    )
    # Create multipolygon

    # Create point
    create_test_file(
        os.path.join(dirpath, "point.geojson"),
        [create_record("point", (1, 1), "Point")],  # Intersects all 4 cells in `grid`
        create_schema("Point")
    )

    # Create points
    create_test_file(
        os.path.join(dirpath, "points.geojson"),
        [
            create_record("point 1", (0.5, 0.5), "Point"),
            create_record("point 2", (1.5, 1.5), "Point")
        ],
        create_schema("Point")
    )

    # Create line
    create_test_file(
        os.path.join(dirpath, "lines.geojson"),
        [
            create_record("A", [(0.5, 0.5), (0.5, 1.5), (1.5, 1.5)], "LineString"),
            create_record("B", [(1, 1), (1.5, 0.5)], "LineString")
        ],
        create_schema("LineString")
    )

    # Create geometry collection
    gc_data = {
        "type": "FeatureCollection",
        "crs": {"type": "name", "properties": {"name": "urn:ogc:def:crs:OGC:1.3:CRS84"}
        },
        "features": [{
            "type": "Feature",
            "geometry": {
                "type": "GeometryCollection",
                "geometries": [{
                    "type": "Polygon",
                    "coordinates": [(0.5, 0.5), (1.5, 0.5), (1.5, 1.5), (0.5, 1.5), (0.5, 0.5)]
                }]
            },
            "properties": {"name": "complicated"}
        }]
    }
    with open("data/gc.geojson", "w") as f:
        json.dump(gc_data, f)

    # Test raster for rasterstats
    array = np.arange(50).reshape((10, 5)).astype(np.float32)
    array[4, :] = -1
    array[5, :] = -1
    create_raster(os.path.join(dirpath, 'range.tif'), array)


if __name__ == '__main__':
    create_test_datasets()



from pandarus import (
    calculate_remaining,
    intersect,
    intersections_from_intersection,
    Map,
    raster_statistics,
)
from pandarus.filesystem import json_importer
from pandarus.calculate import as_features
import fiona
import json
import numpy as np
import os
import pytest
import shutil
import tempfile

dirpath = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))
grid = os.path.join(dirpath, "grid.geojson")
square = os.path.join(dirpath, "square.geojson")
range_raster = os.path.join(dirpath, "range.tif")
dem = os.path.join(dirpath, "DEM.tif")
outside = os.path.join(dirpath, "outside.geojson")
remain_result = os.path.join(dirpath, "remaining.geojson")
inter_res = os.path.join(dirpath, "intersection_result.geojson")
inter_res_md = os.path.join(dirpath, "intersection_result.json.bz2")
inter_res_decompressed = os.path.join(dirpath, "i_result.json")


def fake_zonal_stats(vector, *args, **kwargs):
    for i, f in enumerate(Map(vector, 'name')):
        yield i

def fake_intersection(first, second, indices=None, cpus=None, log_dir=None):
    _, geom = next(Map(second).iter_latlong())
    return {(0, 0): {'measure': 42, 'geom': geom}}

def test_rasterstats_invalid():
    with pytest.raises(AssertionError):
        raster_statistics(grid, 'name', square)

def test_rasterstats_new_path(monkeypatch):
    monkeypatch.setattr(
        'pandarus.calculate.gen_zonal_stats',
        fake_zonal_stats
    )

    fp = raster_statistics(grid, 'name', range_raster, compress=False)
    assert 'rasterstats' in fp
    assert '.json' in fp
    assert os.path.isfile(fp)
    os.remove(fp)

def test_rasterstats(monkeypatch):
    monkeypatch.setattr(
        'pandarus.calculate.gen_zonal_stats',
        fake_zonal_stats
    )

    with tempfile.TemporaryDirectory() as dirpath:
        fp = os.path.join(dirpath, "test.json")
        result = raster_statistics(grid, 'name', range_raster, output=fp, compress=False)
        assert result == fp

        result = json.load(open(fp))

    expected = [
        ['grid cell 0', 0],
        ['grid cell 1', 1],
        ['grid cell 2', 2],
        ['grid cell 3', 3],
    ]

    assert result['metadata'].keys() == {'vector', 'raster', 'when'}
    assert result['metadata']['vector'].keys() == {'field', 'filename', 'path', 'sha256'}
    assert result['metadata']['raster'].keys() == {'band', 'filename', 'path', 'sha256'}
    assert result['data'] == expected

def test_rasterstats_overwrite_existing(monkeypatch):
    monkeypatch.setattr(
        'pandarus.calculate.gen_zonal_stats',
        fake_zonal_stats
    )

    with tempfile.TemporaryDirectory() as dirpath:
        fp = os.path.join(dirpath, "test.json")

        with open(fp, "w") as f:
            f.write("Original content")

        result = raster_statistics(grid, 'name', range_raster, output=fp, compress=False)

        assert result == fp

        content = open(result).read()
        assert content != 'Original content'

def test_rasterstats_mismatched_crs(monkeypatch):
    monkeypatch.setattr(
        'pandarus.calculate.gen_zonal_stats',
        fake_zonal_stats
    )

    with tempfile.TemporaryDirectory() as dirpath:
        fp = os.path.join(dirpath, "test.json")
        with pytest.warns(UserWarning):
            raster_statistics(grid, 'name', dem, output=fp)

def test_as_features(monkeypatch):
    monkeypatch.setattr(
        'pandarus.calculate.mapping',
        lambda x: x
    )

    expected = {
        'geometry': 'Foo',
        'properties': {
            'id': 0,
            'from_label': 1,
            'to_label': 2,
            'measure': 42
        }
    }
    dct = {(1, 2): {'measure': 42, 'geom': 'Foo'}}
    assert next(as_features(dct)) == expected

def test_intersect(monkeypatch):
    monkeypatch.setattr(
        'pandarus.calculate.intersection_dispatcher',
        fake_intersection
    )

    with tempfile.TemporaryDirectory() as dirpath:
        vector_fp, data_fp = intersect(grid, 'name', square, 'name', dirpath=dirpath, compress=False, cpus=None)

        data = json.load(open(data_fp))
        assert data['data'] == [['grid cell 0', 'single', 42]]
        assert data['metadata'].keys() == {'first', 'second', 'when'}
        assert data['metadata']['first'].keys() == {'field', 'filename', 'path', 'sha256'}
        assert data['metadata']['second'].keys() == {'field', 'filename', 'path', 'sha256'}

        expected = {
            'id': '0',
            'type': 'Feature',
            'geometry': {
                'coordinates': [[(0.5, 0.5), (0.5, 1.5), (1.5, 1.5), (1.5, 0.5), (0.5, 0.5)]],
                'type': 'Polygon'
            },
            'properties': dict([('id', 0), ('to_label', 'single'), ('from_label', 'grid cell 0'), ('measure', 42.0)])
        }
        assert next(iter(fiona.open(vector_fp))) == expected

def test_intersect_default_path(monkeypatch):
    monkeypatch.setattr(
        'pandarus.calculate.intersection_dispatcher',
        fake_intersection
    )

    vector_fp, data_fp = intersect(grid, 'name', square, 'name', cpus=None)

    print(vector_fp)
    print(data_fp)

    assert os.path.isfile(vector_fp)
    os.remove(vector_fp)
    assert os.path.isfile(data_fp)
    os.remove(data_fp)

def test_intersect_overwrite_existing(monkeypatch):
    monkeypatch.setattr(
        'pandarus.calculate.intersection_dispatcher',
        fake_intersection
    )

    with tempfile.TemporaryDirectory() as dirpath:
        vector_fp, data_fp = intersect(grid, 'name', square, 'name', dirpath=dirpath, compress=False, cpus=None)

        with open(vector_fp, "w") as f:
            f.write("Weeeee!")
        with open(data_fp, "w") as f:
            f.write("Wooooo!")

        vector_fp, data_fp = intersect(grid, 'name', square, 'name', dirpath=dirpath, compress=False, cpus=None)

        data = json.load(open(data_fp))
        assert data['data'] == [['grid cell 0', 'single', 42]]

        assert len(fiona.open(vector_fp)) == 1

def test_calculate_remaining():
    # Remaining area is 0.5°  by 1°.
    # Circumference of earth is 40.000 km
    # We are close to equator
    # So area should be 1/2 * (4e7 / 360) ** 2 m2
    area = 1/2 * (4e7 / 360) ** 2

    with tempfile.TemporaryDirectory() as dirpath:
        data_fp = calculate_remaining(outside, 'name', remain_result, dirpath=dirpath, compress=False)
        data = json.load(open(data_fp))

    assert data['data'][0][0] == 'by-myself'
    assert np.isclose(area, data['data'][0][1], rtol=1e-2)
    assert data['metadata'].keys() == {'intersections', 'source', 'when'}
    assert data['metadata']['intersections'].keys() == {'field', 'filename', 'path', 'sha256'}
    assert data['metadata']['source'].keys() == {'field', 'filename', 'path', 'sha256'}

def test_calculate_remaining_copmressed_fp():
    with tempfile.TemporaryDirectory() as dirpath:
        data_fp = calculate_remaining(outside, 'name', remain_result, dirpath=dirpath, compress=False)
        assert os.path.isfile(data_fp)

def test_calculate_remaining_default_path():
    data_fp = calculate_remaining(outside, 'name', remain_result, compress=False)
    assert 'intersections' in data_fp
    os.remove(data_fp)

def test_intersections_from_intersection():
    with tempfile.TemporaryDirectory() as dirpath:
        fp1, fp2 = intersections_from_intersection(inter_res, dirpath=dirpath)
        data = json_importer(fp1)
        result = [
            # Order from geojson file
            [0, 'grid cell 3', 3097248058.207057],
            [1, 'grid cell 2', 3097719886.041353],
            [2, 'grid cell 0', 3097719886.0413523],
            [3, 'grid cell 1', 3097248058.207055],
        ]
        assert data['data'] == result
        assert data['metadata'].keys() == {'first', 'second', 'when'}
        assert data['metadata']['first'].keys() == {'field', 'filename', 'path', 'sha256'}
        assert data['metadata']['second'].keys() == {'field', 'filename', 'path', 'sha256'}

        data = json_importer(fp2)
        result = [
            [0, 'single', 3097248058.207057],
            [1, 'single', 3097719886.041353],
            [2, 'single', 3097719886.0413523],
            [3, 'single', 3097248058.207055],
        ]
        assert data['data'] == result
        assert data['metadata'].keys() == {'first', 'second', 'when'}
        assert data['metadata']['first'].keys() == {'field', 'filename', 'path', 'sha256'}
        assert data['metadata']['second'].keys() == {'field', 'filename', 'path', 'sha256'}

def test_intersections_from_intersection_default_path():
    f1, f2 = intersections_from_intersection(inter_res)
    assert 'intersections' in f1
    os.remove(f1)
    os.remove(f2)

def test_intersections_from_intersection_specify_md():
    with tempfile.TemporaryDirectory() as dirpath:
        fp1, _ = intersections_from_intersection(inter_res, inter_res_md, dirpath=dirpath)
        data = json_importer(fp1)
        result = [
            # Order from geojson file
            [0, 'grid cell 3', 3097248058.207057],
            [1, 'grid cell 2', 3097719886.041353],
            [2, 'grid cell 0', 3097719886.0413523],
            [3, 'grid cell 1', 3097248058.207055],
        ]
        assert data['data'] == result

    with tempfile.TemporaryDirectory() as dirpath:
        fp1, _ = intersections_from_intersection(inter_res, inter_res_decompressed, dirpath=dirpath)
        data = json_importer(fp1)
        result = [
            # Order from geojson file
            [0, 'grid cell 3', 3097248058.207057],
            [1, 'grid cell 2', 3097719886.041353],
            [2, 'grid cell 0', 3097719886.0413523],
            [3, 'grid cell 1', 3097248058.207055],
        ]
        assert data['data'] == result

    with tempfile.TemporaryDirectory() as dirpath:
        shutil.copy(inter_res, dirpath)
        new_fp = os.path.join(dirpath, "intersection_result.geojson")

        fp1, _ = intersections_from_intersection(new_fp, inter_res_decompressed, dirpath=dirpath)
        data = json_importer(fp1)
        result = [
            # Order from geojson file
            [0, 'grid cell 3', 3097248058.207057],
            [1, 'grid cell 2', 3097719886.041353],
            [2, 'grid cell 0', 3097719886.0413523],
            [3, 'grid cell 1', 3097248058.207055],
        ]
        assert data['data'] == result

def test_intersections_from_intersection_not_filepath():
    with pytest.raises(AssertionError):
        intersections_from_intersection('')

    with tempfile.TemporaryDirectory() as dirpath:
        shutil.copy(inter_res, dirpath)
        new_fp = os.path.join(dirpath, "intersection_result.geojson")
        with pytest.raises(ValueError):
            intersections_from_intersection(new_fp, dirpath=dirpath)

def test_intersections_from_intersection_find_metadata():
    with tempfile.TemporaryDirectory() as dirpath:
        shutil.copy(inter_res, dirpath)
        new_fp = os.path.join(dirpath, "intersection_result.geojson")

        shutil.copy(inter_res_md, dirpath)

        fp1, _ = intersections_from_intersection(new_fp, dirpath=dirpath)
        data = json_importer(fp1)
        result = [
            # Order from geojson file
            [0, 'grid cell 3', 3097248058.207057],
            [1, 'grid cell 2', 3097719886.041353],
            [2, 'grid cell 0', 3097719886.0413523],
            [3, 'grid cell 1', 3097248058.207055],
        ]
        assert data['data'] == result

    with tempfile.TemporaryDirectory() as dirpath:
        shutil.copy(inter_res, dirpath)
        new_fp = os.path.join(dirpath, "intersection_result.geojson")

        shutil.copy(
            inter_res_decompressed,
            os.path.join(dirpath, 'intersection_result.json')
        )

        fp1, _ = intersections_from_intersection(new_fp, dirpath=dirpath)
        data = json_importer(fp1)
        result = [
            # Order from geojson file
            [0, 'grid cell 3', 3097248058.207057],
            [1, 'grid cell 2', 3097719886.041353],
            [2, 'grid cell 0', 3097719886.0413523],
            [3, 'grid cell 1', 3097248058.207055],
        ]
        assert data['data'] == result



from pandarus.maps import Map, DuplicateFieldID
from rtree import Rtree
import fiona
import os
import pandarus
import pytest


dirpath = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))
grid = os.path.join(dirpath, "grid.geojson")
duplicates = os.path.join(dirpath, "duplicates.geojson")
raster = os.path.join(dirpath, "test_raster_cfs.tif")
countries = os.path.join(dirpath, "test_countries.gpkg")


def test_init():
    m = Map(grid, 'name')
    assert m.filepath == grid
    assert m.file


def test_raster_error(monkeypatch):
    with pytest.raises(AssertionError):
        m = Map(raster, None)

def test_metadata(monkeypatch):
    m = Map(grid, 'name')
    assert m.metadata == {}

    def fake_open(filepath, **others):
        return others

    monkeypatch.setattr(
        pandarus.maps,
        'check_type',
        lambda x: 'vector'
    )
    monkeypatch.setattr(
        pandarus.maps.fiona,
        'open',
        fake_open
    )

    m = Map(grid, 'name', foo='bar')
    assert m.metadata == {'foo': 'bar'}
    assert m.file == {'foo': 'bar'}


def test_get_fieldnames_dictionary():
    m = Map(grid, 'name')
    expected = {0: 'grid cell 0', 1: 'grid cell 1',
                2: 'grid cell 2', 3: 'grid cell 3'}
    assert m.get_fieldnames_dictionary("name") == expected


def test_get_fieldnames_dictionary_errors():
    m = Map(grid, 'name')
    assert m.get_fieldnames_dictionary()
    assert m.get_fieldnames_dictionary(None)
    assert m.get_fieldnames_dictionary("")
    with pytest.raises(AssertionError):
        m.get_fieldnames_dictionary("bar")

    dupes = Map(duplicates, 'name')
    with pytest.raises(DuplicateFieldID):
        dupes.get_fieldnames_dictionary()


def test_properties():
    m = Map(grid, 'name')
    assert m.geometry == 'Polygon'
    assert m.hash
    assert m.crs == '+init=epsg:4326'


def test_magic_methods():
    m = Map(grid, 'name')

    for i, x in enumerate(m):
        pass

    assert i == 3

    expected = {
        'geometry': {
            'type': 'Polygon',
            'coordinates': [[(1.0, 0.0), (1.0, 1.0), (2.0, 1.0), (2.0, 0.0), (1.0, 0.0)]]
            },
        'properties': {'name': 'grid cell 2'},
        'id': '2',
        'type': 'Feature'
    }

    assert m[2] == expected

    assert len(m) == 4

def test_getitem():
    print("Supported Fiona drivers:")
    print(fiona.supported_drivers)

    m = Map(grid, 'name')

    expected = {
        'geometry': {
            'type': 'Polygon',
            'coordinates': [[(1.0, 0.0), (1.0, 1.0), (2.0, 1.0), (2.0, 0.0), (1.0, 0.0)]]
            },
        'properties': {'name': 'grid cell 2'},
        'id': '2',
        'type': 'Feature'
    }
    assert m[2] == expected
    assert hasattr(m, "_index_map")

@pytest.mark.skipif('TRAVIS' in os.environ,
                    reason="No GPKG driver in Travis")
def test_getitem_geopackage():
    print("Supported Fiona drivers:")
    print(fiona.supported_drivers)

    m = Map(countries, 'name')
    assert m[0]
    assert m[0]['id'] == '1'
    assert hasattr(m, "_index_map")

def test_rtree():
    m = Map(grid, 'name')
    r = m.create_rtree_index()
    assert r == m.rtree_index
    assert isinstance(r, Rtree)
